/* auxilliary.pl */

/* This file contains auxilliary predicates for various
   purposes 
*/


%%% member(?Call,+Code)
%%% Call is a member of code Code
member_code(H,(H,_)).
member_code(H,(_,T)) :- !,
	member_code(H,T).
member_code(H,H).

%%% reverse_code(+Code,-ReversedCode,and|or)
%%% reverses the order of the calls in and/or code
reverse_code((H,T),C2,and) :- !,
	reverse_code(T,C2,H,and).
reverse_code((H;T),C2,or) :- 
	not((H = (_->_))), !,
	reverse_code(T,C2,H,or).
reverse_code(H,H,_).

reverse_code((H,T),Result,Temp,and) :- !,
	reverse_code(T,Result,(H,Temp),and).
reverse_code(H,(H,Temp),Temp,and).
reverse_code((H;T),Result,Temp,or) :- 
	not((H = (_->_))), !,
	reverse_code(T,Result,(H;Temp),or).
reverse_code(H,(H;Temp),Temp,or).

%%% append_code(+Code1,+Code2,-Code,+and|or)
%%% concatenates two pieces of and/or code into one
append_code((not(H1),T1),C2,(not(H1);T3),and) :- !,
	append_code(T1,C2,T3,and).
append_code(not(H1),C2,(not(H1);C2),and) :- !.
append_code((H1,T1),C2,(H1,T3),and) :- !,
	append_code(T1,C2,T3,and).
append_code(H1,C2,(H1,C2),and) :- !.
append_code((not(H1);T1),C2,(not(H1),T3),or) :- 
	not((H1 = (_->_))), !,
	append_code(T1,C2,T3,or).
append_code(not(H1),C2,(not(H1),C2),or) :- !.
append_code((H1;T1),C2,(H1;T3),or) :- 
	not((H1 = (_->_))), !,
	append_code(T1,C2,T3,or).
append_code(H1,C2,(H1;C2),or).

%%% replace_code(+OldCall,+NewCall,+OldCode,-NewCode)
%%% replaces the OldCall with the NewCall inside the OldCode
%%% returning the NewCode
replace_code(Old1,New,(Old2,OldTail),(New,NewTail)) :- 
	non_unify_equal(Old1,Old2), !,
	replace_code(Old1,New,OldTail,NewTail).
replace_code(Old,New,(Head,OldTail),(Head,NewTail)) :- !,
	replace_code(Old,New,OldTail,NewTail).
replace_code(Old1,New,Old2,New) :- 
	non_unify_equal(Old1,Old2), !.
replace_code(_,_,Head,Head).

%%% remove_code(+Call,+Code,-NewCode)
%%% removes the Call from the Code returning NewCode
remove_code(Call1,(Call2,Rest),Rest) :- 
	non_unify_equal(Call1,Call2), !.
remove_code(Call,(If >>> Then),(If1 >>> Then1)) :- !,
	remove_code(Call,If,If1),
	remove_code(Call,Then,Then1).
remove_code(Call,(If->Then;Else),(If1->Then1;Else1)) :- !,
	remove_code(Call,If,If1),
	remove_code(Call,Then,Then1),
	remove_code(Call,Else,Else1).
remove_code(Call1,(OneButLast,Call2),OneButLast) :- 
	non_unify_equal(Call1,Call2), !.
remove_code(Call,(First,Rest),(First,Rest1)) :- !,
	remove_code(Call,Rest,Rest1).
remove_code(Call,Last,Last) :- 
	Call \== Last, !.
remove_code(Call1,Call2,_) :- 
	non_unify_equal(Call1,Call2).

%%% unique_code(+Code,-UniqueCode)
%%% removes duplicate calls from Code returning UniqueCode
unique_code((First,Rest),(First,Rest2)) :- 
	member_code(First,Rest),
	remove_code(First,Rest,Rest1),
	nonvar(Rest1), !,
	unique_code(Rest1,Rest2).
unique_code((First,Rest),First) :- 
	member_code(First,Rest),
	remove_code(First,Rest,Rest1),
	var(Rest1), !.
unique_code((First,Rest),(First,Rest1)) :- 
	not(member_code(First,Rest)), !,
	unique_code(Rest,Rest1).
unique_code(Last,Last).

%%% code_length(+Code,-Length)
%%% returns the length of Code (i.e. number of calls)
code_length(Code,Length) :- 
	code_length(Code,Length,0).

code_length((_,B),Length,Temp) :- !,
	Next is Temp + 1,
	code_length(B,Length,Next).
code_length(_,Length,Temp) :- 
	Length is Temp + 1.

%%% subterm(+Term1,+Term2)
%%% checks whether Term1 is a subterm of Term2
%%% by unifying them.  Deterministic search.
subterm(T1,T2) :- 
	nonvar(T1),
	nonvar(T2),
	T1 = T2, !.
subterm(T1,T2) :- 
	var(T1),
	var(T2),
	T1 = T2, !.
subterm(T1,T2) :- 
	nonvar(T2),              % SME: 05/02/93
	functor(T2,_,F),
	F > 0,
	T2 =.. List,
	subterm_aux(T1,List).

%%% non_det_subterm(+Term1,+Term2)
%%% checks whether Term1 is a subterm of Term2
%%% by unifying them.  Non-deterministic search.
non_det_subterm(T1,T2) :- 
	nonvar(T1),
	nonvar(T2),
	T1 = T2.
non_det_subterm(T1,T2) :- 
	var(T1),
	var(T2),
	T1 = T2.
non_det_subterm(T1,T2) :- 
	functor(T2,_,F),
	F > 0,
	T2 =.. List,
	subterm_aux(T1,List).

%%% Auxiliary predicate for subterm
subterm_aux(_,[]) :- !, fail.
subterm_aux(T1,[Head|_]) :- 
	subterm(T1,Head).
subterm_aux(T1,[_|Tail]) :- 
	subterm_aux(T1,Tail).

%%% replace_subterm(+Term1,+Term2,+Term3,-Term4)
%%% checks whether Term1 is a subterm of Term2
%%% If yes then it replaces Term1 with Term3 and
%%% returns Term4
%%% If no then it returns Term3 as Term4
replace_subterm(T1,T2,T3,T3) :- 
	nonvar(T1),
	nonvar(T2),
	T1 = T2, !.
replace_subterm(T1,T2,_,T2) :- 
	nonvar(T1),
	nonvar(T2),
	functor(T1,_,F1),
	functor(T2,_,F2),
	F1 = 0, F2 = 0, !.
replace_subterm(T1,T2,T3,T3) :- 
	var(T1),
	var(T2),
	T1 = T2, !.
replace_subterm(T1,T2,_,T2) :- 
	var(T1),
	var(T2), !.
replace_subterm(T1,T2,T3,T4) :- 
	functor(T2,_,F),
	F > 0,
	T2 =.. List2,
	replace_subterm_aux(T1,List2,T3,List4),
	T4 =.. List4.

replace_subterm_aux(_,[],_,[]) :- !.
replace_subterm_aux(T1,[Head2|Tail2],T3,[Head4|Tail4]) :- 
	replace_subterm(T1,Head2,T3,Head4),
	replace_subterm_aux(T1,Tail2,T3,Tail4).

%%% concat(+Atom1,+Atom2,-Atom3)
%%% This predicate concatenates two atoms.
concat(A1,A2,A) :-
	name(A1,L1),
	name(A2,L2),
	append(L1,L2,L),
	name(A,L).

%%% set_equal(+L1,+L2)
%%% checks if the lists L1 and L2 are equal as sets
set_equal(A,B) :-
	subset(A,B),
	subset(B,A).

%%% difference(+List1,+List2,-List)
%%% This predicate returns the set difference of List2
%%% from List1
difference([],_,[]).
difference([H1|T1],L2,[H1|T]) :-
    not(member(H1,L2)),
    difference(T1,L2,T).
difference([H1|T1],L2,L) :-
    member(H1,L2),
    difference(T1,L2,L).

%%% unique(+List,-UniqueList)
%%% removes the duplicates of List and returns UniqueList
unique([],[]).
unique([H|T],L2) :-
    member(H,T),
    unique(T,L2).
unique([H|T],[H|T2]) :-
    not(member(H,T)),
    unique(T,T2).

%%% substitute(+OldElement,+NewElement,+OldList,-NewList)
%%% replaces the OldElement of the OldList with the NewElement
%%% and returns the NewList
substitute(_,_,[],[]).
substitute(Old1,New,[Old2|Rest],[New|Rest]) :- 
	Old1 == Old2, !.
substitute(Old,New,[Head|T],[Head|T1]) :- 
	substitute(Old,New,T,T1).

%%% member_of_operation(+Var,+Operand,+ListOfOperands)
%%% checks whether there is a slot Operand refering to variable Var
%%% in the ListOfOperands
member_of_operation(Var,H,[H|_]) :- 
	non_det_subterm(slot(_,Var),H),
	atom(Var).
member_of_operation(Var,E,[_|T]) :- 
	member_of_operation(Var,E,T).

%%% remove_vars(+List1,-List2)
%%% removes the unbound variables from List1
%%% returning List2
remove_vars([],[]).
remove_vars([H|T],[H|T1]) :- 
	nonvar(H), !,
	remove_vars(T,T1).
remove_vars([_|T],T1) :- 
	remove_vars(T,T1).

%%% insert(+Call)
%%% calls A.  If A exists and succeeds then ok,
%%% otherwise assert A into database.
insert(A) :- 
	call(A), !.
insert(A) :- 
	asserta(A).

%%% Quintus does not support the use of '<>'
%%% as inequality operand.
A <> B :-
	A \== B.

%%% This predicate abolishes a predicate,
%%% but asserts a sure-to-fail copy of it.
colan_abolish(Predicate,Arity) :-
	abolish(Predicate/Arity),
	functor(Head,Predicate,Arity),
	assertz((Head :- !, fail)).

%%% proves that two terms are unifiable, without unification
%%% to occur
non_unify_equal(A,B) :- 
	nonvar(A), nonvar(B),
	not(not(A=B)).

