/* colanaux.pl */

/* Auxilliary predicates for the CoLan */

/* This file contains general purpose auxiliary predicates
   for code generation and various other utilities
   concerning ADAM
*/


:- dynamic constraint_inheritance_pending/3, call_counter/1,
		variable_binding/2.


%%% check_constraints(new,+Class,?Id,+Attributes)
%%% checks the constraints when a new instance of Class
%%% is created with Attributes.  Id is the identifier of
%%% the instance, but it is non-bound.  It's used only for
%%% unification purposes.
check_constraints(new,Class,Id,Attributes) :- 
	inherit_slot(Class,new_constraints,Code,ConstId),
	execute_code(new,Id,Attributes,Code),
	error_message(new,ConstId),
	!, fail.
check_constraints(_,_,_,_).

%%% check_constraints(delete,+Class,+Id)
%%% checks the constraints when an instance of Class
%%% is about to be deleted.
check_constraints(delete,Class,Id) :- 
	inherit_slot(Class,delete_constraints,Code,ConstId),
	unify_error(delete,Code,ConstId),
	execute_code(delete,Id,Code),
	!, fail.
check_constraints(_,_,_).

%%% check_constraints(+Method,+Slot,+Msg,+Template,+Head)
%%% checks the constraint for Method which is of type Msg
%%% for the Slot.
check_constraints(Method,Slot,Msg,Template,Head) :- 
	message_recipient(Id),
	((get_instance_of(C) => Id,
	  inherit_slot(C,Msg,(Head :- Body),ConstId),
	  unify_error(Head,Body,ConstId),
	  (call(Body) -> fail;(!,fail)));
	 (create_self_send(Method,Slot,Id,Template,Assign),
	  call(Assign))).

%%% checks if a class has some constraints that have been translated
%%% before the class was created and if yes it stores it.
inherit_pending_constraints(Class) :-
	retract(constraint_inheritance_pending(Method,Class,Class1,Constraint)),
	not(inherits_from(Class,Class1)),
	message_name(Method,Name),
	concat(Name,'_constraints',Msg),
	create_message_head(put_,Msg,[Constraint],Call),
	call((Call => Class)),
	fail.
inherit_pending_constraints(_).

%%% inherit_slot(+Class,+Slot,-Value,-Info)
%%% inherits the SLot from Class and all its superclasses.
%%% This predicate was supposed to be general but for the purposes
%%% of CoLan it was only used for inheriting constraints, so it is 
%%% customised for it.  Constraint is returned to value and some
%%% other info about the constraint is returned in Info.
inherit_slot(Class,Slot,Value,constraint(ClassOfConstraint,ConstId)) :-
	get_all_super_classes(Class,Supers),
	bagof(Super,
		(member(Super,[Class|Supers]),
		 create_get_call(Slot,Super,constraint(ConstId,Value),Call),
		 call(Call),
		 ClassOfConstraint=Super),
	      _).

%% This predicate checks whether a certain slot has given a value
%% during the creation of a new instance.
slot_attribute_given(Slot,Class,Attributes,Value) :- 
	slot_type(Slot,Class,Type), !,
	slot_attribute_given_aux(Slot,Attributes,Value,Type).
slot_attribute_given(Slot,_,Attributes,Value) :-
	slot_attribute_given_aux(Slot,Attributes,Value,_).

slot_attribute_given_aux(_,[],_,_) :- !, fail.
slot_attribute_given_aux(Slot,[H|_],Value,single) :-
	H =.. [Slot,[Value]], !.
slot_attribute_given_aux(Slot,[H|_],Value,set) :-
	H =.. [Slot,Value], !.
slot_attribute_given_aux(Slot,[_|T],Value,Type) :-
	slot_attribute_given_aux(Slot,T,Value,Type).

%%% Returns the type of the Slot of Class (set or single)
slot_type(Slot,Class,Type) :-
	get_all_super_classes(Class,Supers),
	bagof(Super,(member(Super,[Class|Supers]),
		     get_slot_desc(slot_desc_tuple(Slot,_,_,Type,_,_)) => Super),
	      _), !.
slot_type(Slot,Class,Type) :-
	exists_class(Class),
	get_instance_of(Super) => Class,
	A = slot_desc_tuple(Slot,_,_,Type,_,_),
	get_slot_desc(A) => Super.

%%% parse_constraint(+Class,+Constraint,-TrimString,-Parse_tree,-Uses)
%%% It parses the Constraint defined at Class.  It returns the 
%%% TrimString which is the constraint string with no funny characters,
%%% the Parse_tree and the uses of the constraint (i.e. early classification)
%%% (see section 3.5.1.)
parse_constraint(Class,Constraint,TrimString,Parse_tree,Uses) :-
	transform(Constraint,ConList),
	re_transform(ConList,TrimString),
	abolish(bonded/2),
	constraint(Class,Parse_tree1,ConList,[]), !,
	number_of_variables(Parse_tree1,2,Parse_tree2),
	optimize_tree(Parse_tree2,Parse_tree),
	usage(Parse_tree,Uses),
	colan_abolish(bonded,2), !.
parse_constraint(_,Constraint,_,[],[]) :- 
	trim(Constraint,TrimString),
	nl,
	write('Syntax error!'),
	nl, nl,
	write('The expression:'), nl,
	write(TrimString), nl,
	write('is not a proper CoLan expression!'),
	nl, nl.

%%% Checks if the constraint has more than Number variables
number_of_variables(constraint(Vars,_),Number,[]) :- 
	length(Vars,Length),
	Length > Number, !,
	nl,
	write('Sorry, but the CoLan expression you have entered'), nl,
	write('cannot be handled by the current system, because'), nl,
	write('it has '), write(Length),
	write(' variables, while the system can handle a maximum'), nl,
	write('of '), write(Number),
	write(' variables.'),
	nl, nl.
number_of_variables(P,_,P).

%%% creates (get_<SLot> => Class) calls
%%% and returns also the Value for unification
%%% purposes
create_get_call(Slot,Class,Value,(Head => Class)) :-
	concat(get_,Slot,Get_call),
	Head =.. [Get_call,Value].

%%% creates messages <Message>_<Slot>
%%% and returns also the Var for unification
%%% purposes
create_message_head(Message,Slot,Var,Call) :-
	concat(Message,Slot,Get_call),
	Call =.. [Get_call,Var].

%%% It prints the appropriate error message
error_message(class_existence,Class) :-
	nonvar(Class), !,
	nl,
	write('The class "'),
	write(Class),
	write('" does not exist!'), nl.
error_message(constraint_existence,constraint(Class,Constraint)) :- !,
	nl,
	write('The constraint: '),
	nl, nl,
	write(Constraint),
	nl, nl,
	write('does not exist either in class "'),
	write(Class),
	write('", or anywhere else in the current schema.'), nl.
error_message(constraint_existence,constraint(Class,ConstId,Constraint)) :- !,
	get_classes(SecondClass) => ConstId,
	nl,
	write('The constraint: '),
	nl, nl,
	write(Constraint),
	nl, nl,
	write('does not exist in class "'),
	write(Class),
	write('", but in class "'), 
	write(SecondClass),
	write('".'), nl.
error_message(Message,constraint(Class,ConstId)) :- 
	get_constraint(String) => ConstId,
	nl,
	write('Error!'), nl, nl,
	write('The message "'),
	write(Message),
	write('" has violated the following integrity constraint:'),
	nl, nl,
	write(String),
	nl, nl,
	write('which is defined in class "'),
	write(Class),
	write('".'), nl.

%%% isa_universal(+Operands,+Univs,+Class,-Operand)
%%% returns the operand from the list of operands 
%%% that refers to a universally quantified variable 
%%% in the Univs list
isa_universal(Operands,Univs,Class,Operand) :- 
	member_of_operation(Var,Operand,Operands),
	atom(Var),
	member(counter(forall(Var),Var,Class),Univs).
isa_universal(Operands,Univs,Class,instance(Var)) :-
	member(instance(Var),Operands),
	member(counter(forall(Var),Var,Class),Univs).

%%% isa_numerical(+Operands,+Nums,+Class,-Pred,-Operand)
%%% returns the operand from the list of operands 
%%% that refers to a numerically quantified variable 
%%% in the Nums list and the quantifier predicate
%%% of the variable
isa_numerical(Operands,Nums,Class,Pred,Operand) :-
	member_of_operation(Var,Operand,Operands),
	member(counter(Pred,Var,Class),Nums).
isa_numerical(Operands,Nums,Class,Pred,instance(Var)) :-
	member(instance(Var),Operands),
	member(counter(Pred,Var,Class),Nums).

%%% inherits_from(+Class1,+Class2)
%%% succeeds if Class1 can inherit from Class2 through
%%% is_a links.
inherits_from(Class,Class) :- !.
inherits_from(Class1,Class2) :-
	path(Class1,Path),
	member(Class2,Path), !.

%%% not_inheritable_classes(+Class,+Classes1,-Classes2)
%%% checks which of the classes in Classes1 cannot inherit
%%% from Class and returns them in Classes2
not_inheritable_classes(Class,Classes1,Classes2) :-
	findall(C,(member(C,Classes1),not(inherits_from(C,Class))),Classes2).

%%% checks if class C has been checked for existence
checked_class(C) :-
	unchecked_class(C),
	exists_class(C), !,
	retract(unchecked_class(C)).
checked_class(C) :-
	unchecked_class(C),
	!, fail.
checked_class(_).

%%% checks if Class exists
exists_class(Class) :- 
	check_comp_class(Class,_).

%%% connects quantifiers with uses
quantifier_symbol(>,delete).
quantifier_symbol(>=,delete).
quantifier_symbol(<,new).
quantifier_symbol(=<,new).
quantifier_symbol(=,new).
quantifier_symbol(=,delete).

%%% connects methods with uses
method_constraints(delete,delete).
method_constraints(delete,update_).
method_constraints(delete,delete_).
method_constraints(new,new).
method_constraints(new,put_).
method_constraints(new,update_).

%%% connects methods with slots of constraints
message_name(delete_,delete_slot) :- !.
message_name(update_,update) :- !.
message_name(put_,put) :- !.
message_name(Name,Name).

%%% usage(+Parse_tree,-Use)
%%% returns the possible uses of a constraint
%%% See section 3.5.1. for early classification
usage([],[]) :- !.
usage(constraint(_,Preds),[new,delete]) :- 
	subterm(in,Preds);
	subterm(not_in,Preds), !.
usage(constraint(Counters,_),Uses) :-
	member(counter(Pred,_,_),Counters),
	Pred =.. [Delete,_,_],
	findall(Use,quantifier_symbol(Delete,Use),RedUses),
	unique(RedUses,Uses).
usage(_,[new]).

%%% binds code with the environment of execution
%%% and then executes it.
execute_code(new,Id,Attributes,Code) :-
	unify(attributes,Attributes,Code),
	unify(id,Id,Code),
	not(call(Code)), !.

execute_code(delete,Id,Code) :-
	subterm((update_deletion_status(_)=>Id),Code), !,
	not(call(Code)), !.

%%% unification predicates for binding the environment
%%% of execution with the code
unify(attributes,Attributes,Code) :-
	subterm(slot_attribute_given(_,_,Attributes,_),Code), !.
unify(id,Id,Code) :-
	subterm(id_pending(Id),Code), !.
unify(_,_,_).

unify_error(Head,Body,ConstId) :- 
	functor(Head,Message,_),
	subterm(error_message(Message,ConstId),Body).

%%% When more than one slots should be retrieved from the list
%%% of attributes, both of the retrieval calls should be unified
%%% with the environment of execution.
unify_only_attributes(new,Call1,Call2) :- 
	subterm(slot_attribute_given(_,_,Attrib,_),Call1),
	subterm(slot_attribute_given(_,_,Attrib,_),Call2), !.
unify_only_attributes(_,_,_).

%%% Just adds the method name to the generated pieces of code
format_with_method(_,[],[]).
format_with_method(Method,[H|T],[method(Method,H)|T1]) :-
	format_with_method(Method,T,T1).

%%% This predicate checks whether the newly created class has an is_a
%%% slot.  If yes then does nothing.  If not, that means that this
%%% class should be a top level one, i.e. is_a 'object'.  Then the
%%% program attaches the is_a slot to the object, containing the
%%% intermediate class colan_object between this class and the 
%%% object class, in order to bug the behaviour of the delete method.
route(Class,Attributes,[is_a([colan_object])|Attributes]) :-
	not(slot_attribute_given(is_a,Class,Attributes,_)), !.
route(_,Attributes,Attributes).

%%% returns all the supeclasses of Class
get_all_super_classes(Class,Supers) :-
	path(Class,Path),
	reverse(Path,[object,colan_object|RSupers]),
	reverse(RSupers,Supers).

%%% create_method_head(+Method,+Class,+Slot,?Template,-Head)
%%% This predicate creates the head of the method definition
create_method_head(Method,Class,Slot,Template,Head) :-
	exists_class(Class), !,
	concat(Method,Slot,Method_name),
	Desc = method_desc_tuple(_,Method_name,Visibility,Cardinality,
					InputArgTypes,ResultType),
	get_method_desc(Desc) => Class,
	Head =.. [Method_name,Visibility,Cardinality,
			InputArgTypes,ResultType,Template].
create_method_head(Method,_,Slot,Template,Head) :-
	concat(Method,Slot,Method_name),
	Head =.. [Method_name,_,_,_,_,Template].

%%% create_self_send(+Msg,+Slot,+Id,+Template,-Assgnment).
%%% This predicate creates an explicit assignment/update of a slot.
create_self_send(put_,Slot,_,[Value],(Slot <-- Value)).
create_self_send(update_,Slot,_,[Old,New],(Slot <-- Old/New)) :-
	my_nonvar(Old), !.
create_self_send(update_,Slot,Id,[_,New],(Slot <-- RealOld/New)) :- !,
	create_message_head(get_,Slot,RealOld,Head),
	call((Head => Id)).
create_self_send(delete_,Slot,_,[Value],(Slot <-- Value/[])) :-
	my_nonvar(Value), !.
create_self_send(delete_,Slot,_,_,(Slot <-- [])).

%%% These predicates return the name of the slot from the message head
find_slot_name(Method,Method_name,Slot) :-
	name(Method,MethodList),
	name(Method_name,Method_name_list),
	append(MethodList,SlotList,Method_name_list),
	name(Slot,SlotList).

find_slot_name(Method_name,Slot) :- 
	name(Method_name,MList),
	keep_rest(95,MList,Rest),
	name(Slot,Rest).

%%% auxiliary predicate for find_slot_name
keep_rest(H,[H|T],T) :- !.
keep_rest(H,[_|T],T1) :- 
	keep_rest(H,T,T1).

%%% This predicate tests whether an object identifier of CoLan
%%% is bound.
my_nonvar(A#B) :- !,
	nonvar(A),
	nonvar(B).
my_nonvar(A) :-
	nonvar(A).

%%% This predicate is an auxiliary one and it is used to check
%%% whether an instance has been checked for deletion during
%%% the execution of the 'delete' method in the colan_object
%%% meta-class.
checked_for_deletion(Id) :-
	get_deletion_status(Status) => Id,
	Status == yes, !.
checked_for_deletion(Id) :-
	get_deletion_status(Status) => Id,
	Status == no,
	!, fail.
checked_for_deletion(_).

%%% This predicate performs the test implied by the quantifier
%%% of the constraint.  Only the 'exactly' quantifier
%%% needs a special treatment, as explained in the section
%%% 3.5.2.8. of the thesis.
quantifier_test(=(Op1,Op2),Method) :- 
	member(Method,[put_,new]),
	not(Op1 is Op2 + 1), !.
quantifier_test(=(Op1,Op2),Method) :- 
	member(Method,[delete,delete_]),
	not(Op1 is Op2 - 1), !.
quantifier_test(=(Op1,Op2),update_) :- 
	not(Op1 is Op2 + 1),
	not(Op1 is Op2 - 1), !.
quantifier_test(Test,_) :-
	call(Test).

%%% Just strips the template of the message
get_template(Head,Template) :- 
	arg(5,Head,Template).

%%% Checks if Slot from Method_name is defined in Class
is_it_defined(Method_name,Class) :- 
	find_slot_name(Method_name,Slot),
	Desc = slot_desc_tuple(Slot,Class,_,_,_,_),
	get_slot_desc(Desc) => Class.

%%% Checks if Method_name is defined in Class
is_it_method_defined(Method_name,Class) :- 
	Desc = method_desc_tuple(Class,Method_name,_,_,_,_),
	get_method_desc(Desc) => Class.

%%% In a chain of function composition finds the inner slot
find_inner_slot(slot(Slot,Arg),Slot) :- 
	atom(Arg), !.
find_inner_slot(slot(_,Arg),Inner) :- 
	find_inner_slot(Arg,Inner).

%%% In a chain of function composition finds the outer slot
find_outer_slot(slot(Slot,_),Slot).

%%% In a chain of function composition finds the inner class
find_inner_class(Instance,Class) :- 
	variable_binding(Instance,CompClass),
	find_inner_class(CompClass,Class), !.
find_inner_class(Class,Class) :- 
	atom(Class), !.
find_inner_class(slot(_,Arg),Class) :- 
	find_inner_class(Arg,Class).

%%% In a chain of function composition finds the outer class
find_outer_class(Class,Class) :- 
	get(Class) => colan_class, !.
find_outer_class(Instance,Class) :- 
	variable_binding(Instance,CompClass), 
	find_outer_class(CompClass,Class), !.
find_outer_class(Instance,Class) :- 
	atom(Instance), !,
	bonded(Instance,Class).
find_outer_class(slot(Slot,Arg),Class) :- 
	find_outer_class(Arg,Inner),
	get_all_super_classes(Inner,Supers),
	bagof(Super,(member(Super,[Inner|Supers]),
		     get_slot_desc(slot_desc_tuple(Slot,_,_,_,_,Class)) => Super),
		_).

%%% create_set_call(+Class,+Slot,?Variable,?InnerVar,-Call)
%%% checks whether a slot is of type set or single and creates
%%% the appropriate call.  The set slots return a list rather than
%%% an item
create_set_call(Class,Slot,Variable,InnerVar,Call) :- 
	slot_type(Slot,Class,set), !,
	create_message_head(get_,Slot,Var,HeadCall),
	Call = findall(Var,(HeadCall => InnerVar),Variable).
create_set_call(_,Slot,Variable,InnerVar,(HeadCall => InnerVar)) :- 
	create_message_head(get_,Slot,Variable,HeadCall).

%%% create_compound_function(+Method,+Slot,+Class,?Variable,-Call)
%%% creates a chain call to implement function composition
create_compound_function(_,Constant,_,Constant,true) :- 
	constant(Constant), !.
create_compound_function(_,instance(_),position(Variable,_),Variable,true) :- !.
create_compound_function(new,slot(Slot,Arg),Class,Variable,Call) :- 
	atom(Arg), not(atom(Class)), !,
	create_compound_class(new,Class,InnerVar,InnerCall),
	create_set_call(Class,Slot,Variable,InnerVar,OuterCall),
	Call = (InnerCall, OuterCall).
create_compound_function(new,slot(Slot,Arg),Class,Variable,Call) :-
	atom(Arg), !,
	Call = slot_attribute_given(Slot,Class,_,Variable).
create_compound_function(delete,slot(Slot,Arg),position(Id,Class),Variable,Call) :-
	atom(Arg), not(atom(Class)), !,
	create_compound_class(delete,Id,Class,InnerVar,InnerCall),
	find_outer_class(Class,OuterClass),
	create_set_call(OuterClass,Slot,Variable,InnerVar,OuterCall),
	Call = (InnerCall, OuterCall).
create_compound_function(delete,slot(Slot,Arg),position(Id,Class),Variable,Call) :-
	atom(Arg), !,
	create_set_call(Class,Slot,Variable,Id,Call).
create_compound_function(num,slot(Slot,Arg),position(Id,Class),Variable,Call) :-
	atom(Arg), not(atom(Class)), !,
	create_compound_class(num,Id,Class,InnerVar,InnerCall),
	create_set_call(Class,Slot,Variable,InnerVar,OuterCall),
	append_code(InnerCall,OuterCall,Call,and).
create_compound_function(num,slot(Slot,Arg),position(Id,Class),Variable,Call) :-
	atom(Arg), !,
	create_set_call(Class,Slot,Variable,Id,Call).
create_compound_function(Method,slot(Slot,Arg),Position,Variable,Call) :-
	create_compound_function(Method,Arg,Position,InnerVar,InnerCall),
	find_outer_class(Arg,OuterClass),
	create_set_call(OuterClass,Slot,Variable,InnerVar,OuterCall),
	Call = (InnerCall, OuterCall).

create_compound_function(Uses,Method,slot(Slot,Arg),Class,Variable,Head,Template) :-
	atom(Arg), !,
	create_method_head(Method,Class,Slot,Template,Head),
	pick_variable(Uses,Method,Template,Variable).
create_compound_function(Uses,Method,slot(Slot,Arg),Class,Variable,Call,Template) :-
	create_compound_function(Uses,Method,Arg,Class,InnerVar,InnerCall,Template),
	find_outer_class(Arg,OuterClass),
	create_set_call(OuterClass,Slot,Variable,InnerVar,OuterCall),
	Call = (InnerCall, OuterCall).

%%% create_compound_class(+Method,+Class,?Variable,-Call)
%%% creates a chain call to implement function composition of
%%% composite classes
create_compound_class(new,slot(Slot,Class),Variable,Call) :-
	atom(Class), !,
	Call = slot_attribute_given(Slot,Class,_,Variable).
create_compound_class(new,slot(Slot,Arg),Variable,Call) :-
	create_compound_class(new,Arg,InnerVar,InnerCall),
	find_outer_class(Arg,OuterClass),
	create_set_call(OuterClass,Slot,Variable,InnerVar,OuterCall),
	Call = (InnerCall, OuterCall).

create_compound_class(delete,Id,slot(Slot,Class),Variable,Call) :-
	atom(Class), !,
	create_set_call(Class,Slot,Variable,Id,Call).
create_compound_class(delete,Id,slot(Slot,Arg),Variable,Call) :- !,
	create_compound_class(delete,Id,Arg,InnerVar,InnerCall),
	find_outer_class(Arg,OuterClass),
	create_set_call(OuterClass,Slot,Variable,InnerVar,OuterCall),
	Call = (InnerCall, OuterCall).
create_compound_class(num,Id,slot(Slot,Class),Variable,(HeadCall => Id)) :-
	atom(Class), !,
	create_message_head(get_,Slot,Variable,HeadCall).
create_compound_class(num,Id,slot(Slot,Arg),Variable,Call) :- 
	create_compound_class(num,Id,Arg,InnerVar,InnerCall),
	create_message_head(get_,Slot,Variable,HeadCall),
	Call = (InnerCall, (HeadCall => InnerVar)).

%%% create_compound_answer_value(+Class,+Operand,-Value,-Call)
%%% creates code for retrieving values from an instance
%%% The Values can be directly compared in the Local Test
%%% The operand is a function composition
create_compound_answer_value(Class,Operand,Value,Call) :- 
	create_compound_answer_value_aux(Operand,Value,Call1,Class,Id),
	Call = (message_recipient(Id), Call1).

%%% auxiliary predicate for create_compound_answer_value
create_compound_answer_value_aux(slot(Slot,Arg),Value,Call,Class,Id) :- 
	atom(Arg), !,
	create_set_call(Class,Slot,Value,Id,Call).
create_compound_answer_value_aux(slot(Slot,Arg),Value,Call,Class,Id) :- 
	create_compound_answer_value_aux(Arg,InnerValue,InnerCall,Class,Id),
	find_outer_class(Arg,OuterClass),
	create_set_call(OuterClass,Slot,Value,InnerValue,OuterCall),
	append_code(InnerCall,OuterCall,Call,and).

%%% creates a new variable Var2 from Var1
%%% This predicate is used in the tree optimization task
%%% if the Var1 ends in a muber then it increases 
%%% this number by one and creates the Var2
%%% Otherwise it adds number 1 at the end of Var1
create_variable(Var1,Var2) :- 
	name(Var1,List1),
	reverse(List1,[Asc1|Rest1]),
	name(Num1,[Asc1]),
	integer(Num1), !,
	Num2 is Num1 + 1,
	name(Num2,[Asc2]),
	reverse(List2,[Asc2|Rest1]),
	name(Var2,List2), !.
create_variable(Var1,Var2) :- 
	name(Var1,List1),
	append(List1,[49],List2),
	name(Var2,List2).

%%% replace_variable(+Var1,+Var2,+Preds1,-Preds2)
%%% replaces all the references of Var1 in Preds1 with Var2
%%% returning Pred2
replace_variable(_,_,[],[]) :- !.
replace_variable(Var1,Var2,[Pred1|Tail1],[Pred2|Tail2]) :- 
	replace_subterm(Var1,Pred1,Var2,Pred2),
	replace_variable(Var1,Var2,Tail1,Tail2).

%%% replaces all variable references in ClassIn with their bindings
%%% i.e. the classes that they really represent
replace_all_variables(ClassIn,ClassOut) :- 
	findall(bonded(Var,Class),(bonded(Var,Class),atom(Class)),List),
	replace_all_variables_aux(List,ClassIn,ClassOut).

replace_all_variables_aux([],Class,Class).
replace_all_variables_aux([bonded(Var,Class)|Tail],ClassIn,ClassOut) :- 
	replace_subterm(Var,ClassIn,Class,ClassTemp),
	replace_all_variables_aux(Tail,ClassTemp,ClassOut).

%%% gives the length of the list Var, if Var is a list
%%% Otherwise it returns one
colan_length(Var,Length) :-
	length(Var,Length), !.
colan_length(_,1).

%%% extends the equality operator to handle equality
%%% of lists
equal(A,B) :-
	A == B, !.
equal(A,B) :-
	set_equal(A,B).

%%% the negation of the previous predicate
not_equal(A,B) :-
	not(equal(A,B)).

%%% tests whether E is a member of the list L,
%%% or if E is a subet of the list L,
%%% if L is a list, otherwise it performs equality test
in(E,L) :-
	member(E,L), !.
in(E,L) :-
	subset(E,L), !.
in(E,E).

%%% the negation of the previous predicate
not_in(E,L) :-
	not(in(E,L)).

%%% negation(+Operator,-Negation)
%%% This precicate takes as input a relational operator
%%% and returns its negation
negation(equal,not_equal) :- !.
negation(not_equal,equal) :- !.
negation(A,B) :- 
	invert_operation(A,B), !.
negation(A,B) :- 
	invert_operation(B,A), !.

%%% adds a check for class existence where two different
%%% classes are involved in a contraint
add_check_class_template(Class,Class,Code,Code) :- !.
add_check_class_template(_,Class2,CodeIn,CodeOut) :- 
	CodeOut = (checked_class(Class2) -> 
			CodeIn;
			(error_message(class_existence,Class2),fail)).

%%% decides if the value that is going to be checked in the
%%% quantification test is going to be different from the
%%% number of instances that pass the predicate of the constraint
%%% i.e. when we add a new instance that satisfies loccaly the constraint
%%% it means that we should add one to the number of instances that 
%%% satisfy the constraint if the update is going to succeed
decide_on_extra(_,put_,0) :- !.
decide_on_extra([new],update,0) :- !.
decide_on_extra(_,_,1).

%%% create_uni_value(+Uses,+Method,+Operand,?Variable,+Class,-Call,?Template)
%%% generates the local call that will deliver the values needed
%%% for the local test.  Because it is usually the universally quantified
%%% variables that need local tests, so is the name 'uni'
create_uni_value(_,new,slot(Slot,Arg),Variable,Class,Call,_) :- 
	create_compound_function(new,slot(Slot,Arg),Class,Variable,Call), !.
create_uni_value(_,new,instance(_),Variable,_,id_pending(Variable),_) :- !,
	insert(id_pending(Variable)).
create_uni_value(_,delete,slot(Slot,Arg),Variable,Id,Call,Class) :- !,
	create_compound_function(delete,slot(Slot,Arg),position(Id,Class),Variable,Call).
create_uni_value(_,delete,instance(_),Id,Id,true,_) :- !.
create_uni_value(Uses,Method,slot(Slot,Arg),Variable,Class,Call,Template) :- !,
	create_compound_function(Uses,Method,slot(Slot,Arg),Class,Variable,Call,Template), !.
create_uni_value(_,Method,instance(_),_,_,true,_) :- 
	member(Method,[put_,delete_,update_]), !.
create_uni_value(_,_,X,X,_,true,_) :-
	constant(X), !.
create_uni_value(_,delete,Complex_pred,Variable,Id,Call,Class) :- !,
	Complex_pred =.. [Operation,Value1,Value2],
	create_uni_value(_,delete,Value1,Variable1,Id,Call1,Class),
	create_uni_value(_,delete,Value2,Variable2,Id,Call2,Class),
	unify_only_attributes(delete,Call1,Call2),
	Eval_pred =.. [Operation,Variable1,Variable2],
	Call = (Call1, Call2, Variable is Eval_pred).
create_uni_value(Uses,Method,Complex_pred,Variable,Class,Call,_) :- 
	Complex_pred =.. [Operation,Value1,Value2],
	create_uni_value(Uses,Method,Value1,Variable1,Class,Call1,_),
	create_uni_value(Uses,Method,Value2,Variable2,Class,Call2,_),
	unify_only_attributes(Method,Call1,Call2),
	Eval_pred =.. [Operation,Variable1,Variable2],
	Call = (Call1, Call2, Variable is Eval_pred).

%%% create_num_value(+Operand,?Variable,+Class,?Instances,-Call)
%%% generates the global call that will deliver the values needed
%%% for the global test.  Because it is usually the numerically quantified
%%% variables that need global tests, so is the name 'num'
create_num_value(slot(Slot,Arg),Variable,Class,Instances,Call) :-
	create_compound_function(num,slot(Slot,Arg),
					position(Instances,Class),Variable,InnerCall),
	find_inner_class(Class,InnerClass),
	Call = (get(Instances) => InnerClass, InnerCall), !.
create_num_value(instance(_),Instances,Class,Instances,Call) :- 
	Call = (get(Instances) => Class), !.
create_num_value(X,X,_,_,true) :- 
	constant(X), !.
create_num_value(Complex_pred,Variable,Class,Instances,Call) :- 
	Complex_pred =.. [Operation,Value1,Value2],
	create_num_value(Value1,Variable1,Class,Instances,Call1),
	create_num_value(Value2,Variable2,Class,Instances,Call2),
	Eval_pred =.. [Operation,Variable1,Variable2],
	Call = (Call1, Call2, Variable is Eval_pred).

%%% create_answer_value(+Class,+Operand,?Value,-Call)
%%% creates code for retrieving values from an instance
%%% The Values can be directly compared in the Local Test
create_answer_value(Class,slot(Slot,Arg),Value,Call) :- !,
	create_compound_answer_value(Class,slot(Slot,Arg),Value,Call).
create_answer_value(_,Value,Value,true) :-
	constant(Value), !.
create_answer_value(Class,Complex_pred,Value,Call) :- 
	Complex_pred =.. [Operation,Op1,Op2],
	create_answer_value(Class,Op1,Value1,Call1),
	create_answer_value(Class,Op2,Value2,Call2),
	Eval_pred =.. [Operation,Value1,Value2],
	Call = (Call1, Call2, Value is Eval_pred).

%%% modify_local_test(+Uses,+Method,+Template,+Value,+OldTest,-NewTest)
%%% checks if a value for delete_<slot> or update_<slot> has been give
%%% and modifies the local test accordingly
modify_local_test([new,delete],update_,[Old,New],_,OldTest,(OldTest;NewTest)) :-
	not(is_list(New)), my_nonvar(Old), !,
	OldTest =.. [Operation,Old,Op2],
	NewTest =.. [Operation,New,Op2].
modify_local_test([new,delete],update_,[Old,New],Value,OldTest,(ValueTest;NewTest)) :- 
	not(is_list(New)), !,
	OldTest =.. [Operation,Old,Op2],
	ValueTest =.. [Operation,Value,Op2],
	NewTest =.. [Operation,New,Op2].
modify_local_test([delete],update_,[Old,New],_,OldTest,(OldTest,not(NewTest))) :- 
	not(is_list(New)), my_nonvar(Old), !,
	OldTest =.. [Operation,Old,Op2],
	NewTest =.. [Operation,New,Op2].
modify_local_test([delete],update_,[Old,New],Value,OldTest,(ValueTest,not(NewTest))) :- 
	not(is_list(New)), !,
	OldTest =.. [Operation,Old,Op2],
	ValueTest =.. [Operation,Value,Op2],
	NewTest =.. [Operation,New,Op2].
modify_local_test(_,delete_,[Old],Value,OldTest,ValueTest) :-
	not(is_list(Old)), not(my_nonvar(Old)), !,
	OldTest =.. [Operation,Old,Op2],
	ValueTest =.. [Operation,Value,Op2].
modify_local_test(_,_,_,_,C,C).

%%% modify_additional_call(+Uses,+Method,+AdditionalCall1,+AdditionalCall2,
%%%						+Template,-AdditionalCall)
%%% decides if there is a need to retrieve a value that has not 
%%% been given and adds the call to the existing one
modify_additional_call([delete],update_,AdditionalCall,_,[Old,_],AdditionalCall) :-
	my_nonvar(Old), !.
modify_additional_call(_,delete_,AdditionalCall,_,[Old],AdditionalCall) :-
	my_nonvar(Old), !.
modify_additional_call(Uses,Method,AdditionalCall1,AdditionalCall2,_,(AdditionalCall2,AdditionalCall1)) :-
	member(delete,Uses),
	member(Method,[update_,delete_]), !.
modify_additional_call(_,_,AdditionalCall,_,_,AdditionalCall).

%%% modify_length(+Uses,+Method,?NewLength,?Length,+Extra,-LengthCall,+Call)
%%% modifies the value to be checked in the quantifier test according to the
%%% type of method.  It also executes the local tests
modify_length(_,_,Length,Length,_,true,true) :- !.
modify_length([new,delete],update_,NewLength,Length,Extra,(NewLength is Length - Extra),(CallA;_)) :- 
	call(CallA), !.
modify_length([new,delete],update_,NewLength,Length,Extra,(NewLength is Length + Extra),(_;CallB)) :- 
	call(CallB), !.
modify_length(Uses,Method,NewLength,Length,Extra,(NewLength is Length + Extra),Call) :- 
	(member(Method,[new,put_]);
	 (Method == update_, Uses == [new])),
	call(Call), !.
modify_length(Uses,Method,NewLength,Length,Extra,(NewLength is Length - Extra),Call) :- 
	(member(Method,[delete,delete_]);
	 (Method == update_, Uses == [delete])),
	call(Call).

%%% modify_test(+Operation,+Operands,+opvar(Op1,Var1),+opvar(Op2,Var2),-Test)
%%% in the case of arithmetic expressions it modifies the test by removing
%%% all the constants from the test and creates the evaluable expression either
%%% by replacing slots with their actual values or constants by themselves
modify_test(Operation1,Operands,opvar(Op1,Var1),opvar(Op2,Var2),Test) :- 
	member(Op,Operands),
	constant(Op), !,
	delete(Operands,Op,[Complex_operand]),
	Complex_operand =.. [Sign|Ops],
	substitute(Op1,Var1,Ops,Value1),
	substitute(Op2,Var2,Value1,Value2),
	Evaluation =.. [Sign|Value2],
	substitute(Complex_operand,Value,Operands,FinalOperands),
	Test1 =.. [Operation1|FinalOperands],
	Test = (Value is Evaluation, Test1).
modify_test(Operation1,_,opvar(_,Var1),opvar(_,Var2),Test) :- 
	Test =.. [Operation1,Var1,Var2].

%%% modify_call(+Uses,+Method,+Call1,+Call2,-Calls)
%%% it returns one or two codes that respond to two different methods
%%% it performs some optimization
modify_call(Uses,Method,(HeadA :- BodyA),(true :- _),Calls) :-
	remove_code(true,BodyA,NewBodyA),
	split_methods(Uses,Method,(HeadA :- NewBodyA),Calls), !.
modify_call(Uses,Method,(true :- _),(HeadB :- BodyB),Calls) :-
	remove_code(true,BodyB,NewBodyB),
	split_methods(Uses,Method,(HeadB :- NewBodyB),Calls), !.
modify_call(Uses,Method,(HeadA :- BodyA),(HeadB :- BodyB),Calls) :-
	remove_code(true,BodyA,NewBodyA),
	remove_code(true,BodyB,NewBodyB),
	split_methods(Uses,Method,(HeadA :- NewBodyA),CallsA),
	split_methods(Uses,Method,(HeadB :- NewBodyB),CallsB), 
	append(CallsA,CallsB,Calls), !.

%%% some times the original code generation routines
%%% create strange pieces of code that are not Horn clauses,
%%% so they need transformation
split_methods(_,_,(Head :- Body),[(Head :- Body)]) :- 
	not(((Head = (_->_));(Head = (_>>>_));(Head = (_,_)))), !.
split_methods(Uses,Method,(MultiHead :- Body),Calls) :- 
	reverse_code(MultiHead,(Evaluation,Tail),and),
	remove_code(true,Tail,Rest),
	Rest = (_,_), !,
	findall(Call,(member_code(Head,Rest),
		      remove_code(Head,Rest,RestHeads),
		      transform_heads_to_gets(Uses,Method,RestHeads,Id,RestGets),
		      Call = (Head :- 
					(message_recipient(Id),
					 (RestGets, Evaluation) >>> Body))),
		Calls).
split_methods(_,_,(MultiHead :- Body),[Call]) :- 
	reverse_code(MultiHead,(Evaluation,Tail),and),
	remove_code(true,Tail,Head),
	Call = (Head :- 
			Evaluation >>> Body).

%%% auxiliary predicate for the transformation to Horn clauses predicate
transform_heads_to_gets(Uses,Method,(Head,Rest),Id,(GetHead,GetRest)) :- !,
	Head =.. [Method_name,_,_,_,_,Template],
	pick_variable(Uses,Method,Template,Variable),
	find_slot_name(Method,Method_name,Slot),
	create_message_head(get_,Slot,Variable,HeadCall),
	GetHead = (HeadCall => Id),
	transform_heads_to_gets(Rest,GetRest).
transform_heads_to_gets(Uses,Method,Head,Id,GetHead) :- 
	Head =.. [Method_name,_,_,_,_,Template],
	pick_variable(Uses,Method,Template,Variable),
	find_slot_name(Method,Method_name,Slot),
	create_message_head(get_,Slot,Variable,HeadCall),
	GetHead = (HeadCall => Id).

%%% chooses a variable from the template according to the method
%%% and to the check we would like to perform
pick_variable(_,_,X,X) :-
	constant(X), !.
pick_variable(_,put_,[Variable],Variable) :- !.
pick_variable([new],update_,[_,Variable],Variable) :- !.
pick_variable(Uses,update_,[Variable,_],Variable) :- 
	member(delete,Uses), !.
pick_variable(_,delete_,[Variable],Variable).

%%% predicate that is used to delete from all the classes in the list
%% the code generated by the constraint with ConstId
delete_all_codes(_,[]) :- !.
delete_all_codes(ConstId,[Class|Rest]) :-
	findall(A,(get_new_constraints(constraint(ConstId,A)) => Class,
		   delete_new_constraints([constraint(ConstId,A)]) => Class),_),
	findall(A,(get_delete_constraints(constraint(ConstId,A)) => Class,
		   delete_delete_constraints([constraint(ConstId,A)]) => Class),_),
	findall(A,(get_put_constraints(constraint(ConstId,A)) => Class,
		   delete_put_constraints([constraint(ConstId,A)]) => Class),_),
	findall(A,(get_update_constraints(constraint(ConstId,A)) => Class,
		   delete_update_constraints([constraint(ConstId,A)]) => Class),_),
	findall(A,(get_delete_slot_constraints(constraint(ConstId,A)) => Class,
		   delete_delete_slot_constraints([constraint(ConstId,A)]) => Class),_),
	delete_all_codes(ConstId,Rest).

%%% inserts into the Prolog database all the variable bindings
assert_bindings(constraint(Vars,_)) :- !,
	findall(Var,(member(counter(_,Var,Class),Vars),
		     asserta(variable_binding(Var,Class))),_).
assert_bindings([]).

%%% retracts from the Prolog database all the variable bindings
retract_bindings :- 
	abolish(variable_binding/2).

