/* control.pl */

/* This file controls the code generation and storage */


:- dynamic id_pending/1.

%%% create_method(+Class,+String,-ConstId)
%%% creates all the code that corresponds to the constraint String
%%% and for the class Class.  Returns the constraint identifier.
create_method(Class,String,ConstId) :-
	parse_constraint(Class,String,TrimString,Parse_tree,Uses),
	assert_bindings(Parse_tree),
	create_method(Class,TrimString,Parse_tree,Uses,Codes,ConstId),
	retract_bindings,
	store_codes(Class,Codes,ConstId).

%%% create_method(+Class,+String,+Parse_tree,+Uses,-Codes,-ConstId)
%%% auxiliary predicate for creating the code.  It returns a list of codes
%%% that correspond to the constraint whose parse tree is given.
%%% It creates the constraint-object and returns the object identifier
create_method(_,_,[],[],[],_) :- !.
create_method(Class,TrimString,Constraint,Uses,[compiled(Class,Codes)|RestCodes],ConstId) :-
	findall(Method,(member(Use,Uses),
			method_constraints(Use,Method)),RedMethods),
	unique(RedMethods,Methods),
	generate_all(Uses,Methods,Class,Constraint,Codes),
	invert_constraint(Use,Methods,Constraint,RestCodes,NotInheritable), 
	new([ConstId,[
		      constraint([TrimString]),
		      classes([Class|NotInheritable])
		]]) => constraints, !.

%%% generate_all(+Uses,+Methods,+Class,+Parse_tree,-Codes)
%%% it generates optimized code for all the affected methods
generate_all(_,[],_,_,[]).
generate_all(Uses,[Method|RestM],Class,Constraint,AllCodes) :-
	generate(Uses,Method,Class,Constraint,Codes),
	optimizer(Codes,OptimizedCodes),
	format_with_method(Method,OptimizedCodes,FormattedCodes),
	generate_all(Uses,RestM,Class,Constraint,RestC),
	append(FormattedCodes,RestC,AllCodes).

%%% generate(+Uses,+Method,+Class,+Parse_tree,-Codes)
%%% it generates non-optimized code for one affected method
generate(Uses,Method,Class,constraint(Vars,Preds),Codes) :- 
	findall(counter(forall(Var),Var,C),
		member(counter(forall(Var),Var,C),Vars),
		Univs),
	difference(Vars,Univs,Nums),
	generate_method(Uses,Method,Class,Univs,Nums,Preds,Codes).

%%% auxiliary predicate for generating code
%%% iterates through the disjuncted statements
generate_method(Uses,Method,Class,[],Nums,[],Call) :-
	generate_method_for_one_pred(Uses,Method,Class,[],Nums,[],Call).
generate_method(Uses,Method,Class,Univs,Nums,[H],F) :- 
	generate_method_and(Uses,Method,Class,Univs,Nums,H,F).
generate_method(Uses,Method,Class,Univs,Nums,[H|T],Codes) :- 
	generate_method_and(Uses,Method,Class,Univs,Nums,H,F),
	generate_method(Uses,Method,Class,Univs,Nums,T,R),
	transform_codes_or(F,R,Codes).

%%% auxiliary predicate for generating code
%%% iterates through the conjuncted statements
generate_method_and(Uses,Method,Class,Univs,Nums,[H],F) :- 
	generate_method_for_one_pred(Uses,Method,Class,Univs,Nums,H,F).
generate_method_and(Uses,Method,Class,Univs,Nums,[H|T],Codes) :- 
	generate_method_for_one_pred(Uses,Method,Class,Univs,Nums,H,F),
	generate_method_and(Uses,Method,Class,Univs,Nums,T,R),
	transform_codes_and(F,R,Codes).

%%% auxiliary predicates to transform the lists of codes where two
%%% pieces of code were created
transform_codes_or([A,B],[C,D],[(A;C),(B;D)]) :- !.
transform_codes_or([A],[B],[(A;B)]).

transform_codes_and([A,B],[C,D],[(A,C),(B,D)]) :- !.
transform_codes_and([A],[B],[(A,B)]).



%%% generate_method_for_one_pred(+Uses,+Method,+UniversalVars,+NumericalVars,+Predicate,-Call)
%%% The core of code generation.  Classifies the constraints to certain types and creates the 
%%% appropriate code.

%%% Numerically quantified only variables with no predicates
generate_method_for_one_pred(_,new,Class,[],		%% Method new for normal class
				[counter(Quantifier_pred,Var,Class)],[],Call) :- 
	atom(Class), !,
	Quantifier_pred =.. [Operation,Var,Number],
	Quantifier_test =.. [Operation,NewLength,Number],
	Call = [(get_counter(Length) => Class,
		 NewLength is Length + 1,
		 quantifier_test(Quantifier_test,new))].
generate_method_for_one_pred(_,new,Class,[],		%% Method new for complex class
				[counter(Quantifier_pred,Var,CompoundClass)],[],Call) :- !,
	Quantifier_pred =.. [Operation,Var,Number],
	Quantifier_test =.. [Operation,Length,Number],
	create_uni_value(_,new,CompoundClass,Variable,Class,Call1,_),
	Call = [(Call1 >>> 
			(colan_length(Variable,Length),
			 quantifier_test(Quantifier_test,new)))].
generate_method_for_one_pred(_,delete,Class,[],		%% Method delete for normal class
				[counter(Quantifier_pred,Var,Class)],[],Call) :- 
	atom(Class), !,
	Quantifier_pred =.. [Operation,Var,Number],
	Quantifier_test =.. [Operation,NewLength,Number],
	Call = [(get_counter(Length) => Class,
		 NewLength is Length - 1,
		 not(quantifier_test(Quantifier_test,delete)) >>>
			(error_message(_,_),
			 message_recipient(Id),
			 update_deletion_status([yes,no]) => Id,
			 fail))].
%% Methods put_, delete_ for complex classes
generate_method_for_one_pred(Uses,Method,Class,[],
				[counter(Quantifier_pred,Var,CompoundClass)],[],Call) :- 
	not(atom(CompoundClass)),
	member(Method,[put_,delete_]), !,
	create_uni_value(Uses,Method,CompoundClass,_,Class,LocalCall,_),
	Quantifier_pred =.. [Operation,Var,Number],
	Quantifier_test =.. [Operation,NewLength,Number],
	create_answer_value(Class,CompoundClass,Value,AdditionalCall),
	Call =
		[(LocalCall :-
			((AdditionalCall,
			  modify_length(Uses,Method,NewLength,Length,1,LengthCall,true)) >>>
		    		(colan_length(Value,Length),
		     	 	 LengthCall,
		     	 	 not(quantifier_test(Quantifier_test,Method)) >>> 
					(error_message(_,_),fail))))].
generate_method_for_one_pred(_,_,_,[],_,[],[]) :- !.	%% Catch-all rule.
%% Universally quantified only variables
generate_method_for_one_pred(_,new,Class,_,[],Predicate,Call) :- !,	%% Method new
 	Predicate =.. [Operation,Op1,Op2],
	Test =.. [Operation,Var1,Var2],
	create_uni_value(_,new,Op1,Var1,Class,Call1,_),
	create_uni_value(_,new,Op2,Var2,Class,Call2,_),
	append_code(Call1,Call2,AppendedCall,and),
	remove_code(true,AppendedCall,HeadCall),
	Call = [(HeadCall >>> Test)].
generate_method_for_one_pred(_,Method,Class,_,[],Predicate,Call) :- 	%% Methods put_,update_
	member(Method,[put_,update_]),
	Predicate =.. [Operation,Op1,Op2],
	not(member(Operation,[in])), !,
	Test1 =.. [Operation,Variable1,Value2],
	invert_operation(Operation,InvertedOperation),
	Test2 =.. [InvertedOperation,Value1,Variable2],
	create_uni_value([new],Method,Op1,Variable1,Class,Call1,_),
	create_uni_value([new],Method,Op2,Variable2,Class,Call2,_),
	create_answer_value(Class,Op1,Value1,AdditionalCall1),
	create_answer_value(Class,Op2,Value2,AdditionalCall2),
	modify_call([new],Method,
		    (Call1 :-
			((AdditionalCall2, not(Test1)) >>> (error_message(_,_),fail))),
		    (Call2 :-
			((AdditionalCall1, not(Test2)) >>> (error_message(_,_),fail))),
		    Call).
generate_method_for_one_pred(_,Method,Class,_,[],Predicate,Call) :- 	%% Methods delete_, update_
	member(Method,[delete_,update_]),
	Predicate =.. [Operation,Op1,Op2],
	not(member(Operation,[not_in])), !,
	Test =.. [Operation,Var1,Var2],
	create_uni_value([delete],Method,Op2,Var2,Class,Call1,_),
	create_answer_value(Class,Op1,Var1,AdditionalCall),
	Call = 
		[(Call1 :-
			((AdditionalCall, Test) >>> (error_message(_,_),fail)))].
generate_method_for_one_pred(_,_,_,_,[],_,[]) :- !.	%% Catch-all rule
%% Numerically quantified only variables
generate_method_for_one_pred(_,new,Class,[],		%% Method new for normal classes
				[counter(Quantifier_pred,Var,Class)],Predicate,Call) :- 
	atom(Class), !,
	Predicate =.. [Operation1,Op1,Op2],
	TestLocal =.. [Operation1,VarLocal1,VarLocal2],
	TestGlobal =.. [Operation1,VarGlobal1,VarGlobal2],
	create_uni_value(_,new,Op1,VarLocal1,Class,LocalCall1,_),
	create_uni_value(_,new,Op2,VarLocal2,Class,LocalCall2,_),
	create_num_value(Op1,VarGlobal1,Class,Instances,GlobalCall1),
	create_num_value(Op2,VarGlobal2,Class,Instances,GlobalCall2),
	Quantifier_pred =.. [Operation2,Var,Number],
	Quantifier_test =.. [Operation2,NewLength,Number],
	append_code(LocalCall1,LocalCall2,LocalCallTemp,and),
	remove_code(true,LocalCallTemp,LocalCall),
	append_code(GlobalCall1,GlobalCall2,GlobalCallTemp,and),
	unique_code(GlobalCallTemp,GlobalCall),
	Call =
		[((LocalCall, TestLocal) >>>
		    		(findall(Instances,(GlobalCall,TestGlobal),List),
		     		 length(List,Length),
		     		 NewLength is Length + 1,
		     		 quantifier_test(Quantifier_test,new)))].
generate_method_for_one_pred(_,new,Class,[],		%% Method new for complex classes
				[counter(Quantifier_pred,Var,CompoundClass)],Predicate,Call) :- !,
	Predicate =.. [Operation1,Op1,Op2],
	TestGlobal =.. [Operation1,VarGlobal1,VarGlobal2],
	create_uni_value(_,new,CompoundClass,VarLocal,Class,LocalCall,_),
	create_compound_function(num,Op1,position(Instances,Class),VarGlobal1,GlobalCall1),
	create_compound_function(num,Op2,position(Instances,Class),VarGlobal2,GlobalCall2),
	Quantifier_pred =.. [Operation2,Var,Number],
	Quantifier_test =.. [Operation2,NewLength,Number],
	append_code(GlobalCall1,GlobalCall2,GlobalCallTemp1,and),
	unique_code(GlobalCallTemp1,GlobalCallTemp2),
	remove_code(true,GlobalCallTemp2,GlobalCall),
	Call =
		[(LocalCall >>>
		    		(findall(Instances,(in(Instances,VarLocal),GlobalCall,TestGlobal),List),
		     		 length(List,Length),
		     		 NewLength is Length + 1,
		     		 quantifier_test(Quantifier_test,new)))].
generate_method_for_one_pred(_,delete,Class,[],		%% Method delete for normal classes
				[counter(Quantifier_pred,Var,Class)],Predicate,Call) :- 
	atom(Class), !,
	Predicate =.. [Operation1,Op1,Op2],
	TestLocal =.. [Operation1,VarLocal1,VarLocal2],
	TestGlobal =.. [Operation1,VarGlobal1,VarGlobal2],
	create_uni_value(_,delete,Op1,VarLocal1,Id,LocalCall1,Class),
	create_uni_value(_,delete,Op2,VarLocal2,Id,LocalCall2,Class),
	create_num_value(Op1,VarGlobal1,Class,Instances,GlobalCall1),
	create_num_value(Op2,VarGlobal2,Class,Instances,GlobalCall2),
	Quantifier_pred =.. [Operation2,Var,Number],
	Quantifier_test =.. [Operation2,NewLength,Number],
	append_code(LocalCall1,LocalCall2,LocalCallTemp,and),
	remove_code(true,LocalCallTemp,LocalCall),
	append_code(GlobalCall1,GlobalCall2,GlobalCallTemp,and),
	unique_code(GlobalCallTemp,GlobalCall),
	Call = [((LocalCall, TestLocal) >>>
		    	(findall(Instances,(GlobalCall,TestGlobal),List),
		     	 length(List,Length),
		     	 NewLength is Length - 1,
		 	 not(quantifier_test(Quantifier_test,delete)) >>>
				(error_message(_,_),
				 update_deletion_status([yes,no]) => Id,
				 fail)))].
%%% Methods put_,update_,delete_ for normal classes
generate_method_for_one_pred(Uses,Method,Class,[],
				[counter(Quantifier_pred,Var,Class)],Predicate,Call) :-
	member(Method,[put_,update_,delete_]), 
	atom(Class), !,
	Predicate =.. [Operation1,Op1,Op2],
	invert_operation(Operation1,InvertedOperation1),
	TestLocal1 =.. [Operation1,VarLocal1,ValueLocal2],
	TestLocal2 =.. [InvertedOperation1,ValueLocal1,VarLocal2],
	TestGlobal =.. [Operation1,VarGlobal1,VarGlobal2],
	create_uni_value(Uses,Method,Op1,VarLocal1,Class,LocalCall1,Template1),
	create_uni_value(Uses,Method,Op2,VarLocal2,Class,LocalCall2,Template2),
	create_num_value(Op1,VarGlobal1,Class,Instances,GlobalCall1),
	create_num_value(Op2,VarGlobal2,Class,Instances,GlobalCall2),
	Quantifier_pred =.. [Operation2,Var,Number],
	Quantifier_test =.. [Operation2,NewLength,Number],
	create_answer_value(Class,Op1,ValueLocal1,AdditionalCall1),
	create_answer_value(Class,Op2,ValueLocal2,AdditionalCall2),
	append_code(GlobalCall1,GlobalCall2,GlobalCallTemp,and),
	unique_code(GlobalCallTemp,GlobalCall),
	modify_call(Uses,Method,
		(LocalCall1 :-
			modify_additional_call(Uses,Method,AdditionalCall2,AdditionalCall1,Template1,AdditionalCall),
			modify_local_test(Uses,Method,Template1,ValueLocal1,TestLocal1,FinalTestLocal1),
			(AdditionalCall, 
			 modify_length(Uses,Method,NewLength,Length,1,LengthCall,FinalTestLocal1)) >>>
		    	(findall(Instances,(GlobalCall,TestGlobal),List),
		     	 length(List,Length),
		     	 LengthCall,
		     	 not(quantifier_test(Quantifier_test,Method)) >>> 
				(error_message(_,_),fail))),
		(LocalCall2 :-
			modify_additional_call(Uses,Method,AdditionalCall1,AdditionalCall2,Template2,AdditionalCall),
			modify_local_test(Uses,Method,Template2,ValueLocal2,TestLocal2,FinalTestLocal2),
			(AdditionalCall, 
			 modify_length(Uses,Method,NewLength,Length,1,LengthCall,FinalTestLocal2)) >>>
		    	(findall(Instances,(GlobalCall,TestGlobal),List),
		     	 length(List,Length),
		     	 LengthCall,
		     	 not(quantifier_test(Quantifier_test,Method)) >>> 
				(error_message(_,_),fail))),
		Call).
%%% Methods put_,update_,delete_ for complex classes
generate_method_for_one_pred(Uses,Method,Class,[],
				[counter(Quantifier_pred,Var,CompoundClass)],Predicate,Call) :-
	member(Method,[put_,update_,delete_]), !,
	Predicate =.. [Operation1,Op1,Op2],
	TestLocal =.. [Operation1,VarLocal1,VarLocal2],
	TestGlobal =.. [Operation1,VarGlobal1,VarGlobal2],
	create_uni_value(Uses,Method,CompoundClass,InnerVar,Class,LocalCall,Template),
	create_compound_function(num,Op1,position(InnerVar,Class),VarLocal1,AdditionalCall11),
	create_compound_function(num,Op2,position(InnerVar,Class),VarLocal2,AdditionalCall12),
	create_compound_class(num,Id,CompoundClass,Instances,InnerCall),
	create_compound_class(num,Id,CompoundClass,InnerVar,AdditionalCall2),
	create_compound_function(num,Op1,position(Instances,Class),VarGlobal1,GlobalCall1),
	create_compound_function(num,Op2,position(Instances,Class),VarGlobal2,GlobalCall2),
	Quantifier_pred =.. [Operation2,Var,Number],
	Quantifier_test =.. [Operation2,NewLength,Number],
	append_code(GlobalCall1,GlobalCall2,GlobalCallTemp1,and),
	unique_code(GlobalCallTemp1,GlobalCallTemp2),
	remove_code(true,GlobalCallTemp2,GlobalCallOuter),
	GlobalCall = (InnerCall, GlobalCallOuter),
	append_code(AdditionalCall11,AdditionalCall12,AdditionalCallTemp1,and),
	unique_code(AdditionalCallTemp1,AdditionalCallTemp2),
	remove_code(true,AdditionalCallTemp2,AdditionalCall1),
	Call = 
		[(LocalCall :- 
			message_recipient(Id),
			modify_additional_call(Uses,Method,AdditionalCall1,AdditionalCall2,Template,AdditionalCall),
			modify_local_test(Uses,Method,Template,InnerVar,TestLocal,FinalTestLocal),
			(AdditionalCall,
			 modify_length(Uses,Method,NewLength,Length,1,LengthCall,FinalTestLocal)) >>>
		    		(findall(Instances,(GlobalCall,TestGlobal),List),
		     	 	 length(List,Length),
		     	 	 LengthCall,
		     	 	 not(quantifier_test(Quantifier_test,Method)) >>> 
					(error_message(_,_),fail)))].
%%% Both numerically and universally quantified variables
%%% Method new for normal second class
generate_method_for_one_pred(_,new,Class1,Univs,Nums,Predicate,[CodeOut]) :- 
	Predicate =.. [Operation1|Operands],
	isa_universal(Operands,Univs,Class1,UnivOp),
	isa_numerical(Operands,Nums,Class2,Quantifier_pred,NumOp),
	atom(Class2), !,
	create_uni_value(_,new,UnivOp,Var1,Class1,Call1,_),
	create_num_value(NumOp,Var2,Class2,Instances,Call2),
	modify_test(Operation1,Operands,opvar(UnivOp,Var1),opvar(NumOp,Var2),Test),
	Quantifier_pred =.. [Operation2,_,Number],
	Quantifier_test =.. [Operation2,Length,Number],
	CodeIn = (Call1 >>>
		    (findall(Instances,(Call2,Test),List),
		     length(List,Length),
		     quantifier_test(Quantifier_test,new))),
	add_check_class_template(Class1,Class2,CodeIn,CodeOut).
%%% Method new for complex second class
generate_method_for_one_pred(_,new,Class1,Univs,Nums,Predicate,[CodeOut]) :- !,
	Predicate =.. [Operation1|Operands],
	isa_universal(Operands,Univs,Class1,UnivOp),
	isa_numerical(Operands,Nums,CompoundClass2,Quantifier_pred,NumOp),
	create_uni_value(_,new,UnivOp,Var1,Class1,InnerCall1,_),
	create_uni_value(_,new,CompoundClass2,InnerVar,Class1,InnerCall2,_),
	unify_only_attributes(new,InnerCall1,InnerCall2),
	create_compound_function(num,NumOp,position(Instances,Class1),Var2,Call2),
	modify_test(Operation1,Operands,opvar(UnivOp,Var1),opvar(NumOp,Var2),Test),
	Quantifier_pred =.. [Operation2,_,Number],
	Quantifier_test =.. [Operation2,Length,Number],
	append_code(InnerCall1,InnerCall2,InnerCall,and),
	CodeIn = (InnerCall >>>
		    (findall(Instances,(in(Instances,InnerVar),Call2,Test),List),
		     length(List,Length),
		     quantifier_test(Quantifier_test,new))),
	add_check_class_template(Class1,CompoundClass2,CodeIn,CodeOut).
%%% Method delete for normal second class
generate_method_for_one_pred(_,delete,Class1,Univs,Nums,Predicate,[CodeOut]) :- 
	Predicate =.. [Operation1|Operands],
	isa_universal(Operands,Univs,Class1,UnivOp),
	isa_numerical(Operands,Nums,Class2,Quantifier_pred,NumOp),
	atom(Class2), !,
	create_uni_value(_,delete,UnivOp,VarLocal1,Id,LocalCall,Class1),
	create_num_value(NumOp,VarGlobal2,Class2,Instances,GlobalCall),
	modify_test(Operation1,Operands,opvar(UnivOp,VarLocal1),
					opvar(NumOp,VarGlobal2),TestGlobal),
	Quantifier_pred =.. [Operation2,_,Number],
	Quantifier_test =.. [Operation2,NewLength,Number],
	CodeIn = (LocalCall >>>
		    	(findall(Instances,
				 (GlobalCall,Instances\==Id,TestGlobal),
				 List),
		     	 length(List,Length),
			 NewLength is Length - 1,
		 	 not(quantifier_test(Quantifier_test,delete)) >>>
				(error_message(_,_),
				 update_deletion_status([yes,no]) => Id,
				 fail))),
	add_check_class_template(Class1,Class2,CodeIn,CodeOut).
%%% Methods put_,update_,delete_ for normal second class
generate_method_for_one_pred(Uses,Method,Class1,Univs,Nums,Predicate,Call) :-
	member(Method,[put_,update_,delete_]),
	Predicate =.. [Operation1|Operands],
	isa_universal(Operands,Univs,Class1,UnivOp),
	isa_numerical(Operands,Nums,Class2,Quantifier_pred,NumOp),
	atom(Class2), !,
	create_uni_value(Uses,Method,UnivOp,VariableIn,Class1,Call1,Template),
	create_uni_value(_,delete,UnivOp,VariableLocal,Id,AdditionalCall,Class1),
	create_num_value(NumOp,Var2,Class2,Instances,Call2),
	modify_test(Operation1,Operands,opvar(UnivOp,VariableIn),
					opvar(NumOp,Var2),Test),
	Quantifier_pred =.. [Operation2,_,Number],
	Quantifier_test =.. [Operation2,NewLength,Number],
	decide_on_extra(Uses,Method,Extra),
	CodeIn = (findall(Instances,(Call2,Instances\==Id,FinalTest),List),
		  length(List,Length),
		  LengthCall,
		  not(quantifier_test(Quantifier_test,Method)) >>> 
			(error_message(_,_),fail)),
	add_check_class_template(Class1,Class2,CodeIn,CodeOut),
	Call = [(Call1 :- 
			message_recipient(Id),
			modify_additional_call(Uses,Method,true,AdditionalCall,Template,AdditionalCall1),
			modify_local_test(Uses,Method,Template,VariableLocal,Test,FinalTest),
			(AdditionalCall1,
			 modify_length(Uses,Method,NewLength,Length,Extra,LengthCall,FinalTest)) >>> 
				CodeOut)].
%%% Methods put_,update_,delete_ for complex second class
generate_method_for_one_pred(Uses,Method,Class1,Univs,Nums,Predicate,Call) :- 
	member(Method,[put_,update_,delete_]), !,
	Predicate =.. [Operation1|Operands],
	isa_universal(Operands,Univs,Class1,UnivOp),
	isa_numerical(Operands,Nums,CompoundClass2,Quantifier_pred,NumOp),
	create_uni_value(Uses,Method,UnivOp,VariableIn1,Class1,Call1,Template1),
	create_uni_value(Uses,Method,CompoundClass2,VariableIn2,Class1,Call2,Template2),
	create_uni_value(_,delete,UnivOp,VariableLocal1,Id,AdditionalCall12,Class1),
	create_compound_class(num,Id,CompoundClass2,VariableIn2,AdditionalCall22),
	create_compound_class(num,Id,CompoundClass2,Instances,InnerCall1),
	create_compound_function(num,NumOp,position(Instances,Class1),GlobalVar21,OuterCall1),
	Call21 = (InnerCall1, OuterCall1),
	create_compound_function(num,NumOp,position(VariableIn2,Class1),LocalVar21,OuterAdditionalCall2),
	AdditionalCall21 = (AdditionalCall12, OuterAdditionalCall2),
	modify_test(Operation1,Operands,opvar(UnivOp,VariableIn1),
					opvar(NumOp,GlobalVar21),Test1),
	modify_test(Operation1,Operands,opvar(UnivOp,VariableLocal1),
					opvar(NumOp,GlobalVar21),TestGlobal2),
	modify_test(Operation1,Operands,opvar(UnivOp,VariableLocal1),
					opvar(NumOp,LocalVar21),TestLocal2),
	Quantifier_pred =.. [Operation2,_,Number],
	Quantifier_test =.. [Operation2,NewLength,Number],
	decide_on_extra(Uses,Method,Extra1),
	CodeIn = (findall(Instances,(Call21,FinalTest),List),
		  length(List,Length),
		  LengthCall,
		  not(quantifier_test(Quantifier_test,Method)) >>> 
			(error_message(_,_),fail)),
	add_check_class_template(Class1,CompoundClass2,CodeIn,CodeOut),
	modify_call(Uses,Method,
		(Call1 :- 
			message_recipient(Id),
			modify_additional_call(Uses,Method,true,AdditionalCall12,Template1,AdditionalCall1),
			modify_local_test(Uses,Method,Template1,VariableLocal1,Test1,FinalTest),
			(AdditionalCall1,
			 modify_length(Uses,Method,NewLength,Length,Extra1,LengthCall,true)) >>> 
				CodeOut),
		(Call2 :- 
			message_recipient(Id),
			modify_additional_call(Uses,Method,AdditionalCall21,AdditionalCall22,Template2,AdditionalCall2),
			modify_local_test(Uses,Method,Template2,VariableLocal1,TestGlobal2,FinalTest),
			(AdditionalCall2,
			 modify_length(Uses,Method,NewLength,Length,1,LengthCall,TestLocal2)) >>> 
				CodeOut),
		Call).
generate_method_for_one_pred(_,_,_,_,_,_,[]).	%% Catch-all rule

